from .cyclic import create_dataset_with_cyclic_features
from .date_features import create_dataset_with_date_features
from .lagged import create_lagged_dataset
from .onehot_dow import create_dataset_with_onehot_dow



def exp_custom(x, terms=10):
    result = 1.0 # x^0 / 0
    numerator = 1.0
    denominator = 1.0

    for n in range(1, terms):
        numerator *= x
        denominator *= n
        result += numerator / denominator

    return result

def dot_custom(a, b):
    return sum(a[i] * b[i] for i in range(len(a)))

def sigmoid(z):
    return 1 / (1 + exp_custom(-z))

def sigmoid_derivative(x):
    s = sigmoid(x)
    return s * (1 - s)


# train.py
import numpy as np
import matplotlib.pyplot as plt
from .network import MLP

class NormalizeFeatures:
    """A class to normalize features of a dataset."""
    def __init__(self, X):
        self.X_mean = X.mean(axis=0)
        self.X_std = X.std(axis=0)
        self.X_std[self.X_std == 0] = 1.0  # Avoid division by zero
        self.X_norm = (X - self.X_mean) / self.X_std

    def normalize(self):
        return self.X_norm
    
    def std(self):
        return self.X_std

    def denormalize(self):
        return self.X_norm * self.X_std + self.X_mean
    
    def mean(self):
        return self.X_mean

def normalize_features(X):
    X_mean = X.mean(axis=0)
    X_std = X.std(axis=0)
    X_std[X_std == 0] = 1.0
    X_norm = (X - X_mean) / X_std
    #X_norm = [np.array(xi).flatten() for xi in X_norm]
    return X_norm, X_mean, X_std

def normalize_targets(y):
    y_min = y.min()
    y_max = y.max()
    y_norm= (y - y_min) / (y_max - y_min)
    y_norm = [float(yi) for yi in y_norm]
    return y_norm

def denormalize_targets(y_norm, y_min, y_max):
    return y_norm * (y_max - y_min) + y_min

def train_model(X_raw, y_raw, hidden_size=10, learning_rate=0.001, epochs=300):
    # Normalize inputs and targets
    X_norm, X_mean, X_std = normalize_features(X_raw)
    y_norm, y_min, y_max = normalize_targets(y_raw)

    # Initialize and train the model
    model = MLP(input_size=X_norm.shape[1], hidden_size=hidden_size, learning_rate=learning_rate)
    model.train(X_norm, y_norm, epochs=epochs)

    return model, X_norm, X_mean, X_std, y_norm, y_min, y_max

def plot_predictions(dates, y_true, y_pred, title="Model Fit"):
    plt.figure(figsize=(12, 5))
    plt.plot(dates, y_true, label="True Sales", marker='o')
    plt.plot(dates, y_pred, label="Predicted Sales", marker='x', linestyle='--')
    plt.title(title)
    plt.xlabel("Date")
    plt.ylabel("Sales")
    plt.xticks(rotation=45)
    plt.grid(True)
    plt.legend()
    plt.tight_layout()
    plt.show()



import numpy as np
import pandas as pd

def create_dataset_with_date_features(df, target_col='sales', lag=5):
    df = df.copy()
    df['sale_date'] = pd.to_datetime(df['sale_date'])
    df['day_of_week'] = df['sale_date'].dt.weekday
    df['day_of_month'] = df['sale_date'].dt.day
    df['month'] = df['sale_date'].dt.month
    df['is_weekend'] = df['day_of_week'].apply(lambda x: 1 if x >= 5 else 0)

    X, y = [], []
    for i in range(lag, len(df)):
        lagged = df[target_col].values[i - lag:i].tolist()
        row = df.iloc[i]
        features = lagged + [
            row['day_of_week'],
            row['day_of_month'],
            row['month'],
            row['is_weekend']
        ]
        X.append(features)
        y.append(row[target_col])
    return np.array(X), np.array(y)


# forecast.py
import numpy as np
import pandas as pd

def sliding_window_forecast(model, X_last, future_dates, X_mean, X_std, y_min, y_max, lag=5):
    """
    Predict N future steps using a sliding window approach.

    Parameters:
    - model: trained MLP model
    - X_last: last normalized input vector used for starting prediction
    - future_dates: list of future pd.Timestamp
    - X_mean/X_std: normalization vectors
    - y_min/y_max: for unnormalizing output
    - lag: number of lagged sales values used

    Returns:
    - list of forecasted actual sales values
    """
    y_preds = []
    current_input = X_last.tolist()

    for date in future_dates:
        y_next_norm = model.forward(current_input)
        y_preds.append(y_next_norm)

        # Slide the lag window
        new_lags = current_input[:lag][1:] + [y_next_norm]  # shift window forward

        # Extract date features
        dow = date.weekday()
        dom = date.day
        month = date.month
        is_weekend = 1 if dow >= 5 else 0

        # Cyclical encoding
        dow_sin = np.sin(2 * np.pi * dow / 7)
        dow_cos = np.cos(2 * np.pi * dow / 7)
        month_sin = np.sin(2 * np.pi * month / 12)
        month_cos = np.cos(2 * np.pi * month / 12)

        # Normalize date features using training mean/std (skip normalization for cyclical features)
        features = [
            dow_sin, dow_cos,
            (dom - X_mean[lag+2]) / X_std[lag+2],
            month_sin, month_cos,
            (is_weekend - X_mean[lag+5]) / X_std[lag+5]
        ]

        current_input = new_lags + features

    # Unnormalize predictions
    return [y * (y_max - y_min) + y_min for y in y_preds]



import numpy as np
import pandas as pd

def create_lagged_dataset(ds, date_col='sale_date', target_col='sales', lag=5):
    """Create a lagged dataset for time series forecasting.
    
    Parameters:
        ds (pd.DataFrame): DataFrame containing the time series data.
        date_col (str): The name of the date column in the DataFrame.
        target_col (str): The name of the target column to forecast.
        lag (int): The number of lagged observations to include.
        
    Returns:
        X (np.ndarray): Array of lagged features.
        y (np.ndarray): Array of target values."""
    X, y = [], []

    ds[date_col] = pd.to_datetime(ds[date_col])
    ds = ds.sort_values(date_col)

    ds['day_of_week'] = ds[date_col].dt.weekday
    ds['is_weekend'] = ds['day_of_week'].isin([5, 6])
    ds['day_of_month'] = ds[date_col].dt.day
    ds['month'] = ds[date_col].dt.month    


    for i in range(lag, len(ds)):
        lagged_values = ds[date_col].values[i - lag:i].tolist()
        row = ds.iloc[i]
        features = lagged_values + [
            row['day_of_week'],
            row['day_of_month'],
            row['month'],
            row['is_weekend']
        ]
        X.append(features)
        y.append(row[target_col])
    return np.array(X), np.array(y)


from .network import Neuron, Layer, MLP
from .preprocessing import create_dataset_with_cyclic_features, create_dataset_with_date_features, create_lagged_dataset, create_dataset_with_onehot_dow
from .utils import exp_custom, dot_custom, sigmoid, sigmoid_derivative
from .train import train_model


# setup.py
from setuptools import setup, find_packages

setup(
    name='araxia',
    version='0.2.0',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'pandas',
        'matplotlib',
    ],
    author='C5m7b4',
    description='A simple from-scratch neural network time series forecaster with preprocessing utilities.',
    url='https://github.com/C5m7b4/Araxia',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
)



import numpy as np
import pandas as pd

def create_dataset_with_cyclic_features(ds, target_col='sales', lag=5):
    """
    Create a dataset with cyclic features for time series forecasting.
    
    Parameters:
    - ds: DataFrame containing the time series data.
    - target_col: Name of the target column to predict.
    - lag: Number of lagged observations to include as features.
    
    Returns:
    - X: Features array with cyclic features and lagged values.
    - y: Target values array.
    """
    ds = ds.copy()
    ds['sale_date'] = pd.to_datetime(ds['sale_date'])
    ds['day_of_week'] = ds['sale_date'].dt.weekday
    ds['day_of_month'] = ds['sale_date'].dt.day
    ds['month'] = ds['sale_date'].dt.month
    ds['is_weekend'] = ds['day_of_week'].apply(lambda x: 1 if x >= 5 else 0)

    ds['dow_sin'] = np.sin(2 * np.pi * ds['day_of_week'] / 7)
    ds['dow_cos'] = np.cos(2 * np.pi * ds['day_of_week'] / 7)
    ds['month_sin'] = np.sin(2 * np.pi * ds['month'] / 12)
    ds['month_cos'] = np.cos(2 * np.pi * ds['month'] / 12)

    X, y = [], []
    for i in range(lag, len(ds)):
        lagged = ds[target_col].values[i - lag:i].tolist()
        row = ds.iloc[i]
        features = lagged + [
            row['dow_sin'], row['dow_cos'],
            row['day_of_week'],
            row['day_of_month'],
            row['month_sin'], row['month_cos'],
            row['is_weekend']
        ]
        X.append(features)
        y.append(row[target_col])
    return np.array(X), np.array(y)



import numpy as np
from .utils import dot_custom, sigmoid, sigmoid_derivative


# Single Neuron
class Neuron:
    """A single neuron with weights and bias, using sigmoid activation."""
    def __init__(self, n_inputs, activation=True):
        # Initialize weights and bias
        self.weights = np.random.randn(n_inputs) * 0.1 # scaled init
        self.bias = 0.0
        self.activation = activation

    def forward(self, x):
        self.last_input = x
        self.last_z = dot_custom(self.weights, x) + self.bias
        self.last_output = sigmoid(self.last_z) if self.activation else self.last_z
        return self.last_output

    def backward(self, dL_dy, learning_rate):
        # Ensure input is 1D
        self.last_input = np.asarray(self.last_input).flatten()

        if self.activation:
            dz = sigmoid_derivative(self.last_z) * dL_dy
        else:
            dz = dL_dy

        dz = float(np.squeeze(dz))

        for i in range(len(self.weights)):
            self.weights[i] -= learning_rate * dz * self.last_input[i]
        self.bias -= learning_rate * dz
    
# Fully Connected Layer
class Layer:
    """A layer of neurons, each neuron can have its own activation function."""
    def __init__(self, n_inputs, n_neurons, activation=True):
        self.neurons = [Neuron(n_inputs, activation=activation) for _ in range(n_neurons)]

    def forward(self, x):
        self.last_input = x
        self.outputs = np.array([neuron.forward(x) for neuron in self.neurons])
        return self.outputs

    def backward(self, dL_dy_list, learning_rate):
        for neuron, dL_dy in zip(self.neurons, dL_dy_list):
            neuron.backward(dL_dy, learning_rate)
    
# Multi-Layer Perceptron with Backprop
class MLP:
    """A simple Multi-Layer Perceptron with one hidden layer and backpropagation.
    
    Paramters:
        - input_size: Number of input features.
        - hidden_size: Number of neurons in the hidden layer.
        - learning_rate: Learning rate for weight updates."""
    def __init__(self, input_size, hidden_size, learning_rate=0.01):
        self.hidden = Layer(input_size, hidden_size,  activation=True)
        self.output = Neuron(hidden_size, activation=False)
        self.lr = learning_rate
        self.loss_history = []

    def forward(self, x):
        self.hidden_output = self.hidden.forward(x)
        return self.output.forward(self.hidden_output)

    def compute_loss(self, y_true, y_pred):
        return (y_true - y_pred) ** 2

    def compute_loss_derivative(self, y_true, y_pred):
        return -2 * (y_true - y_pred)

    def train(self, X, y, epochs=100):
        for epoch in range(epochs):
            total_loss = 0
            for xi, yi in zip(X, y):
                y_pred = self.forward(xi)
                loss = self.compute_loss(yi, y_pred)
                total_loss += loss
                dL_dy = self.compute_loss_derivative(yi, y_pred)
                
                # Backward pass
                self.output.backward(dL_dy, self.lr)
                dL_dh = dL_dy * self.output.weights  # Gradient w.r.t. each hidden neuron output
                self.hidden.backward(dL_dh, self.lr)


            avg_loss = total_loss / len(X)
            self.loss_history.append(avg_loss)

            if epoch % 10 == 0:
                print(f"epoch: {epoch}, | loss: {avg_loss:.6f}")

    def predict(self, X):
        return [self.forward(xi) for xi in X]